
#include <stdlib.h>
#include <avr/io.h>
/*
init5 is after the stack init, but before the c++ class init.
*/
void early_init(void) __attribute__((naked)) __attribute__((section(".init5")));
void early_init(void)
{
	__malloc_margin = 150;
}

#ifndef F_CPU
#define F_CPU 8000000L
#endif

#define BAUD 9600     // 9600 Baud
#define UARTNUM (F_CPU/(BAUD*16l)-1)

void uart_init(void) {
	UBRR0 = UARTNUM;
	UCSR0B = (1<<TXEN0);
}

void printc(char c) {
	while (!(UCSR0A & (1<<UDRE0)));
	UDR0 = c;
}

void print(char * text) {
	while (*text != '\0') {
		printc(*text);
		text++;
	}
}

void printnum(uint16_t val) {
	char buf[6];
	buf[5] = '\0';
	uint8_t i = 0;
	for (i = 0; i < 5; i++) {
		buf[4-i] = (val%10) +'0';
		val /= 10;
	}
	for (i = 0; i < 4; i++) {
		if (buf[i] != '0')
			break;
	}
	print(buf+i);
}

void * s_malloc(uint16_t size) {
	void * m = malloc(size);
	print("m ");
	printnum(size);
	print(" ");
	printnum((uint16_t)m);
	print("\r\n");
	if (m == NULL) {
		print("Out of mem\r\n");
		for(;;);
	}
	if ((uint16_t)m > SP) {
		print("Segfault!\r\n");
		for(;;);
	}
	return m;
}

void s_free(void *ptr) {
	if (ptr != NULL) {
		print("f ");
		printnum((uint16_t)ptr);
		print("\r\n");
		free(ptr);
	}
}

void *s_realloc(void *ptr, uint16_t size) {
	void * m = realloc(ptr, size);
	print("r ");
	printnum((uint16_t)ptr);
	print(" ");
	printnum(size);
	print(" ");
	printnum((uint16_t)m);
	print("\r\n");
	return m;
}

int main(void) {
	uart_init();
	int i;
	void * zero = s_malloc(600);
	s_free(zero);
	for(i = 0; i < 10; i++) {
		void * a = s_malloc(34);
		a = s_realloc(a, 10);
		void * b = s_malloc(34);
		b = s_realloc(b, 12);
		void * c = s_malloc(34);
		c = s_realloc(c, 11);
		s_free(a);
		s_free(b);
		//void * memleak = s_malloc(5); //is 7 bytes per loop
		void * d = s_malloc(60);
		s_free(c);
		void * e = s_malloc(70);
		s_free(d);
		void * f = s_malloc(65);
		s_free(e);
		void * g = s_malloc(80);
		s_free(f);
		void * h = s_malloc(30);
		s_free(g);
		s_free(h);
	}
	print("Test end\r\n");
	for(;;);
}
